/*
/
/ sple_geometry.c - SpatiaLite - spatial extension for SQLite - geometric utility functions
/
/ Author: Sandro Furieri a-furieri@lqt.it
/
/ Copyright (C) 2007  Alessandro Furieri
/
/ This library is free software; you can redistribute it and/or
/ modify it under the terms of the GNU Lesser General Public
/ License as published by the Free Software Foundation; either
/ version 2.1 of the License, or (at your option) any later version.
/
/ This library is distributed in the hope that it will be useful,
/ but WITHOUT ANY WARRANTY; without even the implied warranty of
/ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
/ Lesser General Public License for more details.
/
/ You should have received a copy of the GNU Lesser General Public
/ License along with this library; if not, write to the Free Software
/ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA/
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include "sqlite3ext.h"
#include "spatialite.h"

SPATIALITE_DECLARE sple_geometry_ptr
sple_mbr_decode (const unsigned char *blob, int size)
{
/* 
/ tryes to decode an internal BLOB geometry returning an internal geometric-object 
/
/ this one is a very limited version, that only checks markers/signatures
/ and fetches the MBR, but ignores at all real geometry
*/
    int little_endian;
    sple_geometry_ptr geom = NULL;
    if (size < 45)
	return NULL;		/* cannot be an internal BLOB WKB geometry */
    if (*(blob + 0) != SPLE_MARK_START)
	return NULL;		/* failed to recognize START signature */
    if (*(blob + (size - 1)) != SPLE_MARK_END)
	return NULL;		/* failed to recognize END signature */
    if (*(blob + 38) != SPLE_MARK_MBR)
	return NULL;		/* failed to recognize MBR signature */
    if (*(blob + 1) == SPLE_LITTLE_ENDIAN)
	little_endian = 1;
    else if (*(blob + 1) == SPLE_BIG_ENDIAN)
	little_endian = 0;
    else
	return NULL;		/* unknown encoding; nor litte-endian neither big-endian */
    geom = sple_alloc_geometry ();
    geom->little_endian = little_endian;
    geom->blob = NULL;
    geom->size = 0;
    geom->errors = 0;
    geom->entities = 0;
    geom->geometries = NULL;
    geom->minx = sple_import_64 (blob + 6, little_endian);
    geom->miny = sple_import_64 (blob + 14, little_endian);
    geom->maxx = sple_import_64 (blob + 22, little_endian);
    geom->maxy = sple_import_64 (blob + 30, little_endian);
    return geom;
}

SPATIALITE_DECLARE sple_geometry_ptr
sple_blob_decode (const unsigned char *blob, int size)
{
/* tryes to decode an internal BLOB geometry returning an internal geometric-object */
    int little_endian;
    int type;
    sple_geometry_ptr geom = NULL;
    if (size < 45)
	return NULL;		/* cannot be an internal BLOB WKB geometry */
    if (*(blob + 0) != SPLE_MARK_START)
	return NULL;		/* failed to recognize START signature */
    if (*(blob + (size - 1)) != SPLE_MARK_END)
	return NULL;		/* failed to recognize END signature */
    if (*(blob + 38) != SPLE_MARK_MBR)
	return NULL;		/* failed to recognize MBR signature */
    if (*(blob + 1) == SPLE_LITTLE_ENDIAN)
	little_endian = 1;
    else if (*(blob + 1) == SPLE_BIG_ENDIAN)
	little_endian = 0;
    else
	return NULL;		/* unknown encoding; nor litte-endian neither big-endian */
    geom = sple_alloc_geometry ();
    geom->little_endian = little_endian;
    geom->blob = blob;
    geom->size = size;
    geom->minx = sple_import_64 (blob + 6, little_endian);
    geom->miny = sple_import_64 (blob + 14, little_endian);
    geom->maxx = sple_import_64 (blob + 22, little_endian);
    geom->maxy = sple_import_64 (blob + 30, little_endian);
    type = sple_import_32 (blob + 39, little_endian);
    geom->offset = 43;
    switch (type)
      {
      case SPLE_POINT:
	  sple_decode_point (geom, -1);
	  break;
      case SPLE_LINESTRING:
	  sple_decode_linestring (geom, -1);
	  break;
      case SPLE_POLYGON:
	  sple_decode_polygon (geom, -1);
	  break;
      case SPLE_MULTIPOINT:
      case SPLE_MULTILINESTRING:
      case SPLE_MULTIPOLYGON:
      case SPLE_GEOMETRYCOLLECTION:
	  sple_decode_geometry (geom);
	  break;
      default:
	  geom->errors++;
	  break;
      };
    if (geom->errors)
      {
	  sple_free_geometry (geom);
	  return NULL;
      }
    return geom;
}

SPATIALITE_DECLARE int
sple_get_srid (unsigned char *p_blob)
{
/* gets the SRID for this geometry */
    if (*p_blob != SPLE_MARK_START)
	return -1;
    return sple_import_32 (p_blob + 2, *(p_blob + 1));
}

SPATIALITE_DECLARE void
sple_set_srid (unsigned char *p_blob, int srid)
{
/* sets the SRID for this geometry */
    if (*p_blob != SPLE_MARK_START)
	return;
    sple_export_32 (p_blob + 2, srid, *(p_blob + 1));
}

SPATIALITE_DECLARE void
sple_decode_point (sple_geometry_ptr geom, int position)
{
/* tryes to decode an internal BLOB geometry - POINT */
    sple_point_ptr point;
    double x;
    double y;
    if (geom->size < geom->offset + 16)
      {
	  geom->errors++;
	  return;
      }
    x = sple_import_64 (geom->blob + geom->offset, geom->little_endian);
    y = sple_import_64 (geom->blob + (geom->offset + 8), geom->little_endian);
    geom->offset += 16;
    point = sple_alloc_point (x, y);
    if (position < 0)
      {
	  /* that one is the only entity inside GEOMETRY COLLECTION */
	  geom->entities = 1;
	  geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
	  *(geom->geometries) =
	      sple_alloc_generic_geometry (SPLE_POINT, (void *) point);
      }
    else
      {
	  /* setting the Nth entity within GEOMETRY COLLECTION */
	  if (position >= 0 && position < geom->entities)
	      *(geom->geometries + position) =
		  sple_alloc_generic_geometry (SPLE_POINT, (void *) point);
	  else
	    {
		geom->errors++;
		sple_free_point (point);
	    }
      }
}

SPATIALITE_DECLARE void
sple_decode_linestring (sple_geometry_ptr geom, int position)
{
/* decodes a LINESTRING WKB */
    int vert;
    int iv;
    double x;
    double y;
    sple_linestring_ptr line;
    if (geom->size < geom->offset + 4)
      {
	  geom->errors++;
	  return;
      }
    vert = sple_import_32 (geom->blob + geom->offset, geom->little_endian);
    geom->offset += 4;
    if (geom->size < geom->offset + (16 * vert))
      {
	  geom->errors++;
	  return;
      }
    line = sple_alloc_linestring (vert);
    for (iv = 0; iv < vert; iv++)
      {
	  x = sple_import_64 (geom->blob + geom->offset, geom->little_endian);
	  y = sple_import_64 (geom->blob + (geom->offset + 8),
			      geom->little_endian);
	  sple_set_point (line, iv, x, y);
	  geom->offset += 16;
      }
    if (position < 0)
      {
	  /* that one is the only entity inside GEOMETRY COLLECTION */
	  geom->entities = 1;
	  geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
	  *(geom->geometries) =
	      sple_alloc_generic_geometry (SPLE_LINESTRING, (void *) line);
      }
    else
      {
	  /* setting the Nth entity within GEOMETRY COLLECTION */
	  if (position >= 0 && position < geom->entities)
	      *(geom->geometries + position) =
		  sple_alloc_generic_geometry (SPLE_LINESTRING, (void *) line);
	  else
	    {
		geom->errors++;
		sple_free_linestring (line);
	    }
      }
}

SPATIALITE_DECLARE void
sple_decode_polygon (sple_geometry_ptr geom, int position)
{
/* decodes a POLYGON WKB */
    int borders;
    int vert;
    int iv;
    int ib;
    double x;
    double y;
    sple_polygon_ptr polyg = NULL;
    sple_linestring_ptr ring;
    if (geom->size < geom->offset + 4)
      {
	  geom->errors++;
	  return;
      }
    borders = sple_import_32 (geom->blob + geom->offset, geom->little_endian);
    geom->offset += 4;
    for (ib = 0; ib < borders; ib++)
      {
	  if (geom->size < geom->offset + 4)
	    {
		geom->errors++;
		return;
	    }
	  vert =
	      sple_import_32 (geom->blob + geom->offset, geom->little_endian);
	  geom->offset += 4;
	  if (geom->size < geom->offset + (16 * vert))
	    {
		geom->errors++;
		return;
	    }
	  if (ib == 0)
	    {
		polyg = sple_alloc_polygon (vert, borders - 1);
		ring = polyg->external;
	    }
	  else
	      ring = sple_alloc_internal_border (polyg, ib - 1, vert);
	  for (iv = 0; iv < vert; iv++)
	    {
		x = sple_import_64 (geom->blob + geom->offset,
				    geom->little_endian);
		y = sple_import_64 (geom->blob + (geom->offset + 8),
				    geom->little_endian);
		geom->offset += 16;
		sple_set_point (ring, iv, x, y);
	    }
      }
    if (position < 0)
      {
	  /* that one is the only entity inside GEOMETRY COLLECTION */
	  geom->entities = 1;
	  geom->geometries = malloc (sizeof (sple_generic_geometry_ptr));
	  *(geom->geometries) =
	      sple_alloc_generic_geometry (SPLE_POLYGON, (void *) polyg);
      }
    else
      {
	  /* setting the Nth entity within GEOMETRY COLLECTION */
	  if (position >= 0 && position < geom->entities)
	      *(geom->geometries + position) =
		  sple_alloc_generic_geometry (SPLE_POLYGON, (void *) polyg);
	  else
	    {
		geom->errors++;
		sple_free_polygon (polyg);
	    }
      }
}

SPATIALITE_DECLARE void
sple_decode_geometry (sple_geometry_ptr geom)
{
/* decodes some MULTIxxx or GEOMETRYCOLLECTION WKB */
    int entities;
    int type;
    int ie;
    if (geom->size < geom->offset + 4)
      {
	  geom->errors++;
	  return;
      }
    entities = sple_import_32 (geom->blob + geom->offset, geom->little_endian);
    geom->entities = entities;
    geom->geometries = malloc (sizeof (sple_generic_geometry_ptr) * entities);
    for (ie = 0; ie < entities; ie++)
	*(geom->geometries + ie) = NULL;
    geom->offset += 4;
    for (ie = 0; ie < entities; ie++)
      {
	  if (geom->size < geom->offset + 5)
	    {
		geom->errors++;
		return;
	    }
	  if (*(geom->blob + geom->offset) != SPLE_MARK_ENTITY)
	    {
		/* failed to recognize ENTITY signature */
		geom->errors++;
		return;
	    }
	  type =
	      sple_import_32 (geom->blob + geom->offset + 1,
			      geom->little_endian);
	  geom->offset += 5;
	  switch (type)
	    {
	    case SPLE_POINT:
		sple_decode_point (geom, ie);
		break;
	    case SPLE_LINESTRING:
		sple_decode_linestring (geom, ie);
		break;
	    case SPLE_POLYGON:
		sple_decode_polygon (geom, ie);
		break;
	    default:
		geom->errors++;
		break;
	    };
      }
}

SPATIALITE_DECLARE void
sple_out_wkt (sple_geometry_ptr geom, char **result)
{
/* 
/ prints the WKT representation of current geometry
/ *result* returns the decoded WKT or NULL if any error is encountered
*/
    int txt_size = 1024;
    int ie;
    int type;
    sple_generic_geometry_ptr pGeo;
    *result = malloc (txt_size);
    if (geom->entities == 1)
      {
	  pGeo = *(geom->geometries + 0);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		strcpy (*result, "POINT(");
		sple_out_point ((sple_point_ptr) (pGeo->geometry), result,
				&txt_size);
		sple_out_text (")", result, &txt_size);
		break;
	    case SPLE_LINESTRING:
		strcpy (*result, "LINESTRING(");
		sple_out_linestring ((sple_linestring_ptr) (pGeo->geometry),
				     result, &txt_size);
		sple_out_text (")", result, &txt_size);
		break;
	    case SPLE_POLYGON:
		strcpy (*result, "POLYGON(");
		sple_out_polygon ((sple_polygon_ptr) (pGeo->geometry), result,
				  &txt_size);
		sple_out_text (")", result, &txt_size);
		break;
	    };
      }
    else
      {
	  type = sple_multi_type (geom);
	  switch (type)
	    {
	    case SPLE_POINT:
		strcpy (*result, "MULTIPOINT(");
		for (ie = 0; ie < geom->entities; ie++)
		  {
		      pGeo = *(geom->geometries + ie);
		      if (ie > 0)
			  sple_out_text (", ", result, &txt_size);
		      sple_out_point ((sple_point_ptr) (pGeo->geometry),
				      result, &txt_size);
		  }
		sple_out_text (")", result, &txt_size);
		break;
	    case SPLE_LINESTRING:
		strcpy (*result, "MULTILINESTRING(");
		for (ie = 0; ie < geom->entities; ie++)
		  {
		      pGeo = *(geom->geometries + ie);
		      if (ie > 0)
			  sple_out_text (", (", result, &txt_size);
		      else
			  sple_out_text ("(", result, &txt_size);
		      sple_out_linestring ((sple_linestring_ptr)
					   (pGeo->geometry), result, &txt_size);
		      sple_out_text (")", result, &txt_size);
		  }
		sple_out_text (")", result, &txt_size);
		break;
	    case SPLE_POLYGON:
		strcpy (*result, "MULTIPOLYGON(");
		for (ie = 0; ie < geom->entities; ie++)
		  {
		      pGeo = *(geom->geometries + ie);
		      if (ie > 0)
			  sple_out_text (", (", result, &txt_size);
		      else
			  sple_out_text ("(", result, &txt_size);
		      sple_out_polygon ((sple_polygon_ptr) (pGeo->geometry),
					result, &txt_size);
		      sple_out_text (")", result, &txt_size);
		  }
		sple_out_text (")", result, &txt_size);
		break;
	    default:
		strcpy (*result, "GEOMETRYCOLLECTION(");
		for (ie = 0; ie < geom->entities; ie++)
		  {
		      if (ie > 0)
			  sple_out_text (", ", result, &txt_size);
		      pGeo = *(geom->geometries + ie);
		      switch (pGeo->type)
			{
			case SPLE_POINT:
			    strcat (*result, "POINT(");
			    sple_out_point ((sple_point_ptr) (pGeo->geometry),
					    result, &txt_size);
			    sple_out_text (")", result, &txt_size);
			    break;
			case SPLE_LINESTRING:
			    strcat (*result, "LINESTRING(");
			    sple_out_linestring ((sple_linestring_ptr)
						 (pGeo->geometry), result,
						 &txt_size);
			    sple_out_text (")", result, &txt_size);
			    break;
			case SPLE_POLYGON:
			    strcat (*result, "POLYGON(");
			    sple_out_polygon ((sple_polygon_ptr)
					      (pGeo->geometry), result,
					      &txt_size);
			    sple_out_text (")", result, &txt_size);
			    break;
			};
		      sple_out_text (")", result, &txt_size);
		  }
		break;
	    };
      }
}

SPATIALITE_DECLARE sple_geometry_ptr
sple_alloc_geometry ()
{
/* allocates and initializes a GEOMETRY COLLECTION */
    sple_geometry_ptr p = malloc (sizeof (sple_geometry));
    p->blob = NULL;
    p->size = 0;
    p->errors = 0;
    p->entities = 0;
    p->geometries = NULL;
    p->minx = DBL_MAX;
    p->miny = DBL_MAX;
    p->maxx = DBL_MIN;
    p->maxy = DBL_MIN;
    return p;
}

SPATIALITE_DECLARE void
sple_free_geometry (sple_geometry_ptr p)
{
/* cleans all memory allocations related to this GEOMETRY COLLECTION */
    int i;
    if (!p)
	return;
    for (i = 0; i < p->entities; i++)
      {
	  if (*(p->geometries + i))
	      sple_free_generic_geometry (*(p->geometries + i));
      }
    if (p->geometries)
	free (p->geometries);
    free (p);
}

SPATIALITE_DECLARE sple_point_ptr
sple_alloc_point (double x, double y)
{
/* allocates and initializes a POINT */
    sple_point_ptr p = malloc (sizeof (sple_point));
    p->x = x;
    p->y = y;
    return p;
}

SPATIALITE_DECLARE void
sple_free_point (sple_point_ptr p)
{
/* cleans all memory allocations related to this POINT */
    if (!p)
	return;
    free (p);
}

SPATIALITE_DECLARE sple_linestring_ptr
sple_alloc_linestring (int vert)
{
/* allocates a LINESTRING */
    if (vert <= 0)
	return NULL;
    sple_linestring_ptr p = malloc (sizeof (sple_linestring));
    p->points = vert;
    if (vert <= 0)
	p->geometry = NULL;
    else
	p->geometry = malloc (sizeof (sple_point) * vert);
    p->minx = DBL_MAX;
    p->miny = DBL_MAX;
    p->maxx = DBL_MIN;
    p->maxy = DBL_MIN;
    return p;
}

SPATIALITE_DECLARE void
sple_free_linestring (sple_linestring_ptr p)
{
/* cleans all memory allocations related to this LINESTRING */
    if (!p)
	return;
    if (p->geometry)
	free (p->geometry);
    free (p);
}

SPATIALITE_DECLARE sple_linestring_ptr
sple_alloc_internal_border (sple_polygon_ptr polyg, int position, int vert)
{
/* allocates a POLYGON's internal border */
    if (position >= 0 && position < polyg->borders)
	;
    else
	return NULL;
    if (vert <= 0)
	return NULL;
    sple_linestring_ptr p = malloc (sizeof (sple_linestring));
    *(polyg->internals + position) = p;
    p->points = vert;
    if (vert <= 0)
	p->geometry = NULL;
    else
	p->geometry = malloc (sizeof (sple_point) * vert);
    p->minx = DBL_MAX;
    p->miny = DBL_MAX;
    p->maxx = DBL_MIN;
    p->maxy = DBL_MIN;
    return p;
}

SPATIALITE_DECLARE sple_polygon_ptr
sple_alloc_polygon (int ext_vert, int int_borders)
{
/* allocates a POLYGON */
    int i;
    if (ext_vert <= 0)
	return NULL;
    sple_polygon_ptr p = malloc (sizeof (sple_polygon));
    p->external = sple_alloc_linestring (ext_vert);
    p->borders = int_borders;
    if (int_borders <= 0)
      {
	  p->borders = 0;
	  p->internals = NULL;
      }
    else
      {
	  p->borders = int_borders;
	  p->internals = malloc (sizeof (sple_linestring_ptr) * int_borders);
	  for (i = 0; i < int_borders; i++)
	      *(p->internals + i) = NULL;
      }
    p->minx = DBL_MAX;
    p->miny = DBL_MAX;
    p->maxx = DBL_MIN;
    p->maxy = DBL_MIN;
    return p;
}

SPATIALITE_DECLARE void
sple_free_polygon (sple_polygon_ptr p)
{
/* cleans all memory allocations related to this POLYGON */
    int i;
    if (!p)
	return;
    if (p->external)
	sple_free_linestring (p->external);
    for (i = 0; i < p->borders; i++)
	sple_free_linestring (*(p->internals + i));
    if (p->internals)
	free (p->internals);
    free (p);
}

SPATIALITE_DECLARE sple_generic_geometry_ptr
sple_alloc_generic_geometry (short type, void *ptr)
{

/* allocates and initializes a generic geometry container */
    sple_generic_geometry_ptr p = malloc (sizeof (sple_generic_geometry));
    p->type = type;
    p->geometry = ptr;
    return p;
}

SPATIALITE_DECLARE void
sple_free_generic_geometry (sple_generic_geometry_ptr p)
{
/* cleans all memory allocations related to this generic geometry */
    switch (p->type)
      {
      case SPLE_POINT:
	  sple_free_point ((sple_point_ptr) (p->geometry));
	  break;
      case SPLE_LINESTRING:
	  sple_free_linestring ((sple_linestring_ptr) (p->geometry));
	  break;
      case SPLE_POLYGON:
	  sple_free_polygon ((sple_polygon_ptr) (p->geometry));
	  break;
      };
}

SPATIALITE_DECLARE void
sple_set_point (sple_linestring_ptr p, int vert, double x, double y)
{
/* sets the coordinates for a LINESTRING's vertex */
    if (p == NULL)
	return;
    if (vert >= 0 && vert < p->points)
      {
	  (p->geometry + vert)->x = x;
	  (p->geometry + vert)->y = y;
      }
}

SPATIALITE_DECLARE unsigned short
sple_import_16 (const unsigned char *p, int little_endian)
{
    union cvt
    {
	unsigned char byte[2];
	unsigned short short_value;
    } convert;
#ifdef __LITTLE_ENDIAN__
/* Litte-Endian architecture [e.g. x86] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  convert.byte[0] = *(p + 1);
	  convert.byte[1] = *(p + 0);
      }
    else
      {
	  /* Little Endian data */
	  convert.byte[0] = *(p + 0);
	  convert.byte[1] = *(p + 1);
      }
#else
/* Big Endian architecture [e.g. PPC] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  convert.byte[0] = *(p + 0);
	  convert.byte[1] = *(p + 1);
      }
    else
      {
	  /* Little Endian data */
	  convert.byte[0] = *(p + 1);
	  convert.byte[1] = *(p + 0);
      }
#endif
    return convert.short_value;
}

SPATIALITE_DECLARE unsigned int
sple_import_32 (const unsigned char *p, int little_endian)
{
    union cvt
    {
	unsigned char byte[4];
	unsigned int int_value;
    } convert;
#ifdef __LITTLE_ENDIAN__
/* Litte-Endian architecture [e.g. x86] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  convert.byte[0] = *(p + 3);
	  convert.byte[1] = *(p + 2);
	  convert.byte[2] = *(p + 1);
	  convert.byte[3] = *(p + 0);
      }
    else
      {
	  /* Little Endian data */
	  convert.byte[0] = *(p + 0);
	  convert.byte[1] = *(p + 1);
	  convert.byte[2] = *(p + 2);
	  convert.byte[3] = *(p + 3);
      }
#else
/* Big Endian architecture [e.g. PPC] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  convert.byte[0] = *(p + 0);
	  convert.byte[1] = *(p + 1);
	  convert.byte[2] = *(p + 2);
	  convert.byte[3] = *(p + 3);
      }
    else
      {
	  /* Little Endian data */
	  convert.byte[0] = *(p + 3);
	  convert.byte[1] = *(p + 2);
	  convert.byte[2] = *(p + 1);
	  convert.byte[3] = *(p + 0);
      }
#endif
    return convert.int_value;
}

SPATIALITE_DECLARE double
sple_import_64 (const unsigned char *p, int little_endian)
{
    union cvt
    {
	unsigned char byte[8];
	double double_value;
    } convert;
#ifdef __LITTLE_ENDIAN__
/* Litte-Endian architecture [e.g. x86] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  convert.byte[0] = *(p + 7);
	  convert.byte[1] = *(p + 6);
	  convert.byte[2] = *(p + 5);
	  convert.byte[3] = *(p + 4);
	  convert.byte[4] = *(p + 3);
	  convert.byte[5] = *(p + 2);
	  convert.byte[6] = *(p + 1);
	  convert.byte[7] = *(p + 0);
      }
    else
      {
	  /* Little Endian data */
	  convert.byte[0] = *(p + 0);
	  convert.byte[1] = *(p + 1);
	  convert.byte[2] = *(p + 2);
	  convert.byte[3] = *(p + 3);
	  convert.byte[4] = *(p + 4);
	  convert.byte[5] = *(p + 5);
	  convert.byte[6] = *(p + 6);
	  convert.byte[7] = *(p + 7);
      }
#else
/* Big Endian architecture [e.g. PPC] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  convert.byte[0] = *(p + 0);
	  convert.byte[1] = *(p + 1);
	  convert.byte[2] = *(p + 2);
	  convert.byte[3] = *(p + 3);
	  convert.byte[4] = *(p + 4);
	  convert.byte[5] = *(p + 5);
	  convert.byte[6] = *(p + 6);
	  convert.byte[7] = *(p + 7);
      }
    else
      {
	  /* Little Endian data */
	  convert.byte[0] = *(p + 7);
	  convert.byte[1] = *(p + 6);
	  convert.byte[2] = *(p + 5);
	  convert.byte[3] = *(p + 4);
	  convert.byte[4] = *(p + 3);
	  convert.byte[5] = *(p + 2);
	  convert.byte[6] = *(p + 1);
	  convert.byte[7] = *(p + 0);
      }
#endif
    return convert.double_value;
}

SPATIALITE_DECLARE void
sple_export_16 (unsigned char *p, unsigned short value, int little_endian)
{
    union cvt
    {
	unsigned char byte[2];
	unsigned short short_value;
    } convert;
    convert.short_value = value;
#ifdef __LITTLE_ENDIAN__
/* Litte-Endian architecture [e.g. x86] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  *(p + 0) = convert.byte[1];
	  *(p + 1) = convert.byte[0];
      }
    else
      {
	  /* Little Endian data */
	  *(p + 0) = convert.byte[0];
	  *(p + 1) = convert.byte[1];
      }
#else
/* Big Endian architecture [e.g. PPC] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  *(p + 0) = convert.byte[0];
	  *(p + 1) = convert.byte[1];
      }
    else
      {
	  /* Little Endian data */
	  *(p + 0) = convert.byte[1];
	  *(p + 1) = convert.byte[0];
      }
#endif
}

SPATIALITE_DECLARE void
sple_export_32 (unsigned char *p, unsigned int value, int little_endian)
{
    union cvt
    {
	unsigned char byte[4];
	unsigned long int_value;
    } convert;
    convert.int_value = value;
#ifdef __LITTLE_ENDIAN__
/* Litte-Endian architecture [e.g. x86] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  *(p + 0) = convert.byte[3];
	  *(p + 1) = convert.byte[2];
	  *(p + 2) = convert.byte[1];
	  *(p + 3) = convert.byte[0];
      }
    else
      {
	  /* Little Endian data */
	  *(p + 0) = convert.byte[0];
	  *(p + 1) = convert.byte[1];
	  *(p + 2) = convert.byte[2];
	  *(p + 3) = convert.byte[3];
      }
#else
/* Big Endian architecture [e.g. PPC] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  *(p + 0) = convert.byte[0];
	  *(p + 1) = convert.byte[1];
	  *(p + 2) = convert.byte[2];
	  *(p + 3) = convert.byte[3];
      }
    else
      {
	  /* Little Endian data */
	  *(p + 0) = convert.byte[3];
	  *(p + 1) = convert.byte[2];
	  *(p + 2) = convert.byte[1];
	  *(p + 3) = convert.byte[0];
      }
#endif
}

SPATIALITE_DECLARE void
sple_export_64 (unsigned char *p, double value, int little_endian)
{
    union cvt
    {
	unsigned char byte[8];
	double double_value;
    } convert;
    convert.double_value = value;
#ifdef __LITTLE_ENDIAN__
/* Litte-Endian architecture [e.g. x86] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  *(p + 0) = convert.byte[7];
	  *(p + 1) = convert.byte[6];
	  *(p + 2) = convert.byte[5];
	  *(p + 3) = convert.byte[4];
	  *(p + 4) = convert.byte[3];
	  *(p + 5) = convert.byte[2];
	  *(p + 6) = convert.byte[1];
	  *(p + 7) = convert.byte[0];
      }
    else
      {
	  /* Little Endian data */
	  *(p + 0) = convert.byte[0];
	  *(p + 1) = convert.byte[1];
	  *(p + 2) = convert.byte[2];
	  *(p + 3) = convert.byte[3];
	  *(p + 4) = convert.byte[4];
	  *(p + 5) = convert.byte[5];
	  *(p + 6) = convert.byte[6];
	  *(p + 7) = convert.byte[7];
      }
#else
/* Big Endian architecture [e.g. PPC] */
    if (!little_endian)
      {
	  /* Big Endian data */
	  *(p + 0) = convert.byte[0];
	  *(p + 1) = convert.byte[1];
	  *(p + 2) = convert.byte[2];
	  *(p + 3) = convert.byte[3];
	  *(p + 4) = convert.byte[4];
	  *(p + 5) = convert.byte[5];
	  *(p + 6) = convert.byte[6];
	  *(p + 7) = convert.byte[7];
      }
    else
      {
	  /* Little Endian data */
	  *(p + 0) = convert.byte[7];
	  *(p + 1) = convert.byte[6];
	  *(p + 2) = convert.byte[5];
	  *(p + 3) = convert.byte[4];
	  *(p + 4) = convert.byte[3];
	  *(p + 5) = convert.byte[2];
	  *(p + 6) = convert.byte[1];
	  *(p + 7) = convert.byte[0];
      }
#endif
}

SPATIALITE_DECLARE short
sple_multi_type (sple_geometry_ptr geom)
{
/* checks the type of entities within a collection */
    int ie;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    sple_generic_geometry_ptr pGeo;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		n_points++;
		break;
	    case SPLE_LINESTRING:
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		n_polygons++;
		break;
	    };
      }
    if (n_points > 0 && n_linestrings == 0 && n_polygons == 0)
	return SPLE_POINT;
    if (n_points == 0 && n_linestrings > 0 && n_polygons == 0)
	return SPLE_LINESTRING;
    if (n_points == 0 && n_linestrings == 0 && n_polygons > 0)
	return SPLE_POLYGON;
    return SPLE_UNKNOWN;
}

SPATIALITE_DECLARE void
sple_out_text (char *text, char **buffer, int *size)
{
/* formats a WKT generic text */
    sple_out_check_buffer (buffer, size);
    strcat (*buffer, text);
}

SPATIALITE_DECLARE void
sple_out_point (sple_point_ptr point, char **buffer, int *size)
{
/* formats a WKT POINT */
    char buf_x[128];
    char buf_y[128];
    char buf[256];
    sple_out_check_buffer (buffer, size);
    sprintf (buf_x, "%1.6lf", point->x);
    sple_out_clean (buf_x);
    sprintf (buf_y, "%1.6lf", point->y);
    sple_out_clean (buf_y);
    sprintf (buf, "%s %s", buf_x, buf_y);
    strcat (*buffer, buf);
}

SPATIALITE_DECLARE void
sple_out_linestring (sple_linestring_ptr line, char **buffer, int *size)
{
/* formats a WKT LINESTRING */
    char buf_x[128];
    char buf_y[128];
    char buf[256];
    int iv;
    for (iv = 0; iv < line->points; iv++)
      {
	  sple_out_check_buffer (buffer, size);
	  sprintf (buf_x, "%1.6lf", (line->geometry + iv)->x);
	  sple_out_clean (buf_x);
	  sprintf (buf_y, "%1.6lf", (line->geometry + iv)->y);
	  sple_out_clean (buf_y);
	  if (iv > 0)
	      sprintf (buf, ", %s %s", buf_x, buf_y);
	  else
	      sprintf (buf, "%s %s", buf_x, buf_y);
	  strcat (*buffer, buf);
      }
}

SPATIALITE_DECLARE void
sple_out_polygon (sple_polygon_ptr polyg, char **buffer, int *size)
{
/* formats a WKT POLYGON */
    char buf_x[128];
    char buf_y[128];
    char buf[256];
    int ib;
    int iv;
    sple_linestring_ptr border = polyg->external;
    for (iv = 0; iv < border->points; iv++)
      {
	  sple_out_check_buffer (buffer, size);
	  sprintf (buf_x, "%1.6lf", (border->geometry + iv)->x);
	  sple_out_clean (buf_x);
	  sprintf (buf_y, "%1.6lf", (border->geometry + iv)->y);
	  sple_out_clean (buf_y);
	  if (iv == 0)
	      sprintf (buf, "(%s %s", buf_x, buf_y);
	  else if (iv == (polyg->external->points - 1))
	      sprintf (buf, ", %s %s)", buf_x, buf_y);
	  else
	      sprintf (buf, ", %s %s", buf_x, buf_y);
	  strcat (*buffer, buf);
      }
    for (ib = 0; ib < polyg->borders; ib++)
      {
	  border = *(polyg->internals + ib);
	  for (iv = 0; iv < border->points; iv++)
	    {
		sple_out_check_buffer (buffer, size);
		sprintf (buf_x, "%1.6lf", (border->geometry + iv)->x);
		sple_out_clean (buf_x);
		sprintf (buf_y, "%1.6lf", (border->geometry + iv)->y);
		sple_out_clean (buf_y);
		if (iv == 0)
		    sprintf (buf, ", (%s %s", buf_x, buf_y);
		else if (iv == (border->points - 1))
		    sprintf (buf, ", %s %s)", buf_x, buf_y);
		else
		    sprintf (buf, ", %s %s", buf_x, buf_y);
		strcat (*buffer, buf);
	    }
      }
}

SPATIALITE_DECLARE void
sple_out_check_buffer (char **buffer, int *size)
{
/* checks if the receiving buffer has enough free room, and in case reallocates it */
    char *old = *buffer;
    int len = strlen (*buffer);
    if ((*size - len) < 256)
      {
	  *size += 4096;
	  *buffer = realloc (old, *size);
      }
}

SPATIALITE_DECLARE void
sple_out_clean (char *buffer)
{
/* cleans unneded trailing zeros */
    int i;
    for (i = strlen (buffer) - 1; i >= 0; i--)
      {
	  if (buffer[i] == '0')
	      buffer[i] = '\0';
	  else
	      break;
      }
    if (buffer[i] == '.')
	buffer[i] = '\0';
}

SPATIALITE_DECLARE void
sple_blob_encode (sple_geometry_ptr geom, unsigned char **result, int *size)
{
/* 
/ builds the BLOB representation for this GEOMETRY
 */
    int ie;
    int ib;
    int iv;
    int entities = 0;
    int n_points = 0;
    int n_linestrings = 0;
    int n_polygons = 0;
    int type;
    unsigned char *ptr;
    sple_generic_geometry_ptr pGeo;
    sple_point_ptr point = NULL;
    sple_linestring_ptr line = NULL;
    sple_polygon_ptr polyg = NULL;
    sple_mbr_geometry (geom);
    for (ie = 0; ie < geom->entities; ie++)
      {
	  /* how many entities, and of what kind, do we have ? */
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		point = (sple_point_ptr) (pGeo->geometry);
		entities++;
		n_points++;
		break;
	    case SPLE_LINESTRING:
		line = (sple_linestring_ptr) (pGeo->geometry);
		entities++;
		n_linestrings++;
		break;
	    case SPLE_POLYGON:
		polyg = (sple_polygon_ptr) (pGeo->geometry);
		entities++;
		n_polygons++;
		break;
	    };
      }
/* ok, we can determine the geometry class */
    if (n_points == 1 && n_linestrings == 0 && n_polygons == 0)
	type = SPLE_POINT;
    else if (n_points > 1 && n_linestrings == 0 && n_polygons == 0)
	type = SPLE_MULTIPOINT;
    else if (n_points == 0 && n_linestrings == 1 && n_polygons == 0)
	type = SPLE_LINESTRING;
    else if (n_points == 0 && n_linestrings > 1 && n_polygons == 0)
	type = SPLE_MULTILINESTRING;
    else if (n_points == 0 && n_linestrings == 0 && n_polygons == 1)
	type = SPLE_POLYGON;
    else if (n_points == 0 && n_linestrings == 0 && n_polygons > 1)
	type = SPLE_MULTIPOLYGON;
    else
	type = SPLE_GEOMETRYCOLLECTION;
/* and now we compute the size of BLOB */
    *size = 44;			/* header size */
    if (type == SPLE_POINT)
	*size += (sizeof (double) * 2);	/* two points */
    else if (type == SPLE_LINESTRING)
	*size += (4 + ((sizeof (double) * 2) * line->points));	/* # points + [x,y] for each vertex */
    else if (type == SPLE_POLYGON)
      {
	  line = polyg->external;
	  *size += (8 + ((sizeof (double) * 2) * line->points));	/* # rings + # points + [x.y] array - exterior ring */
	  for (ib = 0; ib < polyg->borders; ib++)
	    {
		line = *(polyg->internals + ib);
		*size += (4 + ((sizeof (double) * 2) * line->points));	/* # points + [x,y] array - interior ring */
	    }
      }
    else
      {
	  /* this one is not a simple geometry; should be a MULTIxxxx or a GEOMETRYCOLLECTION */
	  *size += 4;		/* # entities */
	  for (ie = 0; ie < geom->entities; ie++)
	    {
		*size += 5;	/* entity header */
		pGeo = *(geom->geometries + ie);
		switch (pGeo->type)
		  {
		  case SPLE_POINT:
		      *size += (sizeof (double) * 2);	/* two points */
		      break;
		  case SPLE_LINESTRING:
		      line = (sple_linestring_ptr) (pGeo->geometry);
		      *size += (4 + ((sizeof (double) * 2) * line->points));	/* # points + [x,y] for each vertex */
		      break;
		  case SPLE_POLYGON:
		      polyg = (sple_polygon_ptr) (pGeo->geometry);
		      line = polyg->external;
		      *size += (8 + ((sizeof (double) * 2) * line->points));	/* # rings + # points + [x.y] array - exterior ring */
		      for (ib = 0; ib < polyg->borders; ib++)
			{
			    line = *(polyg->internals + ib);
			    *size += (4 + ((sizeof (double) * 2) * line->points));	/* # points + [x,y] array - interior ring */
			}
		      break;
		  };
	    }
      }
    *result = malloc (*size);
    ptr = *result;
/* and finally we build the BLOB */
    if (type == SPLE_POINT)
      {
	  *ptr = SPLE_MARK_START;	/* START signatue */
	  *(ptr + 1) = SPLE_LITTLE_ENDIAN;	/* byte ordering */
	  sple_export_32 (ptr + 2, -1, 1);	/* the SRID */
	  sple_export_64 (ptr + 6, geom->minx, 1);	/* MBR - minimun X */
	  sple_export_64 (ptr + 14, geom->miny, 1);	/* MBR - minimun Y */
	  sple_export_64 (ptr + 22, geom->maxx, 1);	/* MBR - maximun X */
	  sple_export_64 (ptr + 30, geom->maxy, 1);	/* MBR - maximun Y */
	  *(ptr + 38) = SPLE_MARK_MBR;	/* MBR signature */
	  sple_export_32 (ptr + 39, SPLE_POINT, 1);	/* class POINT */
	  sple_export_64 (ptr + 43, point->x, 1);	/* X */
	  sple_export_64 (ptr + 51, point->y, 1);	/* Y */
	  *(ptr + 59) = SPLE_MARK_END;	/* END signature */
      }
    else if (type == SPLE_LINESTRING)
      {
	  *ptr = SPLE_MARK_START;	/* START signatue */
	  *(ptr + 1) = SPLE_LITTLE_ENDIAN;	/* byte ordering */
	  sple_export_32 (ptr + 2, -1, 1);	/* the SRID */
	  sple_export_64 (ptr + 6, geom->minx, 1);	/* MBR - minimun X */
	  sple_export_64 (ptr + 14, geom->miny, 1);	/* MBR - minimun Y */
	  sple_export_64 (ptr + 22, geom->maxx, 1);	/* MBR - maximun X */
	  sple_export_64 (ptr + 30, geom->maxy, 1);	/* MBR - maximun Y */
	  *(ptr + 38) = SPLE_MARK_MBR;	/* MBR signature */
	  sple_export_32 (ptr + 39, SPLE_LINESTRING, 1);	/* class LINESTRING */
	  sple_export_32 (ptr + 43, line->points, 1);	/* # points */
	  ptr += 47;
	  for (iv = 0; iv < line->points; iv++)
	    {
		sple_export_64 (ptr, (line->geometry + iv)->x, 1);
		sple_export_64 (ptr + 8, (line->geometry + iv)->y, 1);
		ptr += 16;
	    }
	  *ptr = SPLE_MARK_END;	/* END signature */
      }
    else if (type == SPLE_POLYGON)
      {
	  *ptr = SPLE_MARK_START;	/* START signatue */
	  *(ptr + 1) = SPLE_LITTLE_ENDIAN;	/* byte ordering */
	  sple_export_32 (ptr + 2, -1, 1);	/* the SRID */
	  sple_export_64 (ptr + 6, geom->minx, 1);	/* MBR - minimun X */
	  sple_export_64 (ptr + 14, geom->miny, 1);	/* MBR - minimun Y */
	  sple_export_64 (ptr + 22, geom->maxx, 1);	/* MBR - maximun X */
	  sple_export_64 (ptr + 30, geom->maxy, 1);	/* MBR - maximun Y */
	  *(ptr + 38) = SPLE_MARK_MBR;	/* MBR signature */
	  sple_export_32 (ptr + 39, SPLE_POLYGON, 1);	/* class POLYGON */
	  sple_export_32 (ptr + 43, polyg->borders + 1, 1);	/* # rings */
	  line = polyg->external;
	  sple_export_32 (ptr + 47, line->points, 1);	/* # points - exterior ring */
	  ptr += 51;
	  for (iv = 0; iv < line->points; iv++)
	    {
		sple_export_64 (ptr, (line->geometry + iv)->x, 1);	/* X - exterior ring */
		sple_export_64 (ptr + 8, (line->geometry + iv)->y, 1);	/* Y - exterior ring */
		ptr += 16;
	    }
	  for (ib = 0; ib < polyg->borders; ib++)
	    {
		line = *(polyg->internals + ib);
		sple_export_32 (ptr, line->points, 1);	/* # points - interior ring */
		ptr += 4;
		for (iv = 0; iv < line->points; iv++)
		  {
		      sple_export_64 (ptr, (line->geometry + iv)->x, 1);	/* X - interior ring */
		      sple_export_64 (ptr + 8, (line->geometry + iv)->y, 1);	/* Y - interior ring */
		      ptr += 16;
		  }
	    }
	  *ptr = SPLE_MARK_END;	/* END signature */
      }
    else
      {
	  /* this one is a MULTIxxxx or a GEOMETRYCOLLECTION - building the main header */
	  *ptr = SPLE_MARK_START;	/* START signatue */
	  *(ptr + 1) = SPLE_LITTLE_ENDIAN;	/* byte ordering */
	  sple_export_32 (ptr + 2, -1, 1);	/* the SRID */
	  sple_export_64 (ptr + 6, geom->minx, 1);	/* MBR - minimun X */
	  sple_export_64 (ptr + 14, geom->miny, 1);	/* MBR - minimun Y */
	  sple_export_64 (ptr + 22, geom->maxx, 1);	/* MBR - maximun X */
	  sple_export_64 (ptr + 30, geom->maxy, 1);	/* MBR - maximun Y */
	  *(ptr + 38) = SPLE_MARK_MBR;	/* MBR signature */
	  sple_export_32 (ptr + 39, type, 1);	/* geometric class */
	  sple_export_32 (ptr + 43, entities, 1);	/* # entities */
	  ptr += 47;
	  for (ie = 0; ie < geom->entities; ie++)
	    {
		pGeo = *(geom->geometries + ie);
		switch (pGeo->type)
		  {
		  case SPLE_POINT:
		      point = (sple_point_ptr) (pGeo->geometry);
		      *ptr = SPLE_MARK_ENTITY;	/* ENTITY signature */
		      sple_export_32 (ptr + 1, SPLE_POINT, 1);	/* class POINT */
		      sple_export_64 (ptr + 5, point->x, 1);	/* X */
		      sple_export_64 (ptr + 13, point->y, 1);	/* Y */
		      ptr += 21;
		      break;
		  case SPLE_LINESTRING:
		      line = (sple_linestring_ptr) (pGeo->geometry);
		      *ptr = SPLE_MARK_ENTITY;	/* ENTITY signature */
		      sple_export_32 (ptr + 1, SPLE_LINESTRING, 1);	/* class LINESTRING */
		      sple_export_32 (ptr + 5, line->points, 1);	/* # points */
		      ptr += 9;
		      for (iv = 0; iv < line->points; iv++)
			{
			    sple_export_64 (ptr, (line->geometry + iv)->x, 1);	/* X */
			    sple_export_64 (ptr + 8, (line->geometry + iv)->y, 1);	/* Y */
			    ptr += 16;
			}
		      break;
		  case SPLE_POLYGON:
		      polyg = (sple_polygon_ptr) (pGeo->geometry);
		      *ptr = SPLE_MARK_ENTITY;	/* ENTITY signature */
		      sple_export_32 (ptr + 1, SPLE_POLYGON, 1);	/* class POLYGON */
		      sple_export_32 (ptr + 5, polyg->borders + 1, 1);	/* # rings */
		      line = polyg->external;
		      sple_export_32 (ptr + 9, line->points, 1);	/* # points - exterior ring */
		      ptr += 13;
		      for (iv = 0; iv < line->points; iv++)
			{
			    sple_export_64 (ptr, (line->geometry + iv)->x, 1);	/* X - exterior ring */
			    sple_export_64 (ptr + 8, (line->geometry + iv)->y, 1);	/* Y - exterior ring */
			    ptr += 16;
			}
		      for (ib = 0; ib < polyg->borders; ib++)
			{
			    line = *(polyg->internals + ib);
			    sple_export_32 (ptr, line->points, 1);	/* # points - interior ring */
			    ptr += 4;
			    for (iv = 0; iv < line->points; iv++)
			      {
				  sple_export_64 (ptr, (line->geometry + iv)->x, 1);	/* X - interior ring */
				  sple_export_64 (ptr + 8, (line->geometry + iv)->y, 1);	/* Y - interior ring */
				  ptr += 16;
			      }
			}
		      *ptr = SPLE_MARK_END;	/* END signature */
		      break;
		  };
	    }
	  *ptr = SPLE_MARK_END;	/* END signature */
      }
}

SPATIALITE_DECLARE void
sple_shift_coords (sple_geometry_ptr geom, double shift_x, double shift_y,
		   unsigned char **result, int *size)
{
/* 
/ return a geometry that is the old old geometry with required shifting applied to coordinates
 */
    int ie;
    int ib;
    int iv;
    sple_generic_geometry_ptr pGeo;
    sple_point_ptr point;
    sple_polygon_ptr polyg;
    sple_linestring_ptr line;
    *size = 0;
    *result = NULL;
    double minx = DBL_MAX;
    double miny = DBL_MAX;
    double maxx = DBL_MIN;
    double maxy = DBL_MIN;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		point = (sple_point_ptr) (pGeo->geometry);
		point->x += shift_x;
		point->y += shift_y;
		if (point->x < minx)
		    minx = point->x;
		if (point->y < miny)
		    miny = point->y;
		if (point->x > maxx)
		    maxx = point->x;
		if (point->y > maxy)
		    maxy = point->y;
		break;
	    case SPLE_LINESTRING:
		line = (sple_linestring_ptr) (pGeo->geometry);
		for (iv = 0; iv < line->points; iv++)
		  {
		      (line->geometry + iv)->x += shift_x;
		      (line->geometry + iv)->y += shift_y;
		      if ((line->geometry + iv)->x < minx)
			  minx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y < miny)
			  miny = (line->geometry + iv)->y;
		      if ((line->geometry + iv)->x > maxx)
			  maxx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y > maxy)
			  maxy = (line->geometry + iv)->y;
		  }
		break;
	    case SPLE_POLYGON:
		polyg = (sple_polygon_ptr) (pGeo->geometry);
		line = polyg->external;
		for (iv = 0; iv < line->points; iv++)
		  {
		      (line->geometry + iv)->x += shift_x;
		      (line->geometry + iv)->y += shift_y;
		      if ((line->geometry + iv)->x < minx)
			  minx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y < miny)
			  miny = (line->geometry + iv)->y;
		      if ((line->geometry + iv)->x > maxx)
			  maxx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y > maxy)
			  maxy = (line->geometry + iv)->y;
		  }
		for (ib = 0; ib < polyg->borders; ib++)
		  {
		      line = *(polyg->internals + ib);
		      for (iv = 0; iv < line->points; iv++)
			{
			    (line->geometry + iv)->x += shift_x;
			    (line->geometry + iv)->y += shift_y;
			    if ((line->geometry + iv)->x < minx)
				minx = (line->geometry + iv)->x;
			    if ((line->geometry + iv)->y < miny)
				miny = (line->geometry + iv)->y;
			    if ((line->geometry + iv)->x > maxx)
				maxx = (line->geometry + iv)->x;
			    if ((line->geometry + iv)->y > maxy)
				maxy = (line->geometry + iv)->y;
			}
		  }
		break;
	    };
      }
    geom->minx += minx;
    geom->miny += miny;
    geom->maxx += maxx;
    geom->maxy += maxy;
/* builds the modified GEOMETRY to be passed back */
    sple_blob_encode (geom, result, size);	/* generating BLOB internal representation */
}

SPATIALITE_DECLARE void
sple_scale_coords (sple_geometry_ptr geom, double scale_x, double scale_y,
		   unsigned char **result, int *size)
{
/* 
/ return a geometry that is the old old geometry with required scaling applied to coordinates
 */
    int ie;
    int ib;
    int iv;
    sple_generic_geometry_ptr pGeo;
    sple_point_ptr point;
    sple_polygon_ptr polyg;
    sple_linestring_ptr line;
    *size = 0;
    *result = NULL;
    double minx = DBL_MAX;
    double miny = DBL_MAX;
    double maxx = DBL_MIN;
    double maxy = DBL_MIN;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		point = (sple_point_ptr) (pGeo->geometry);
		point->x *= scale_x;
		point->y *= scale_y;
		if (point->x < minx)
		    minx = point->x;
		if (point->y < miny)
		    miny = point->y;
		if (point->x > maxx)
		    maxx = point->x;
		if (point->y > maxy)
		    maxy = point->y;
		break;
	    case SPLE_LINESTRING:
		line = (sple_linestring_ptr) (pGeo->geometry);
		for (iv = 0; iv < line->points; iv++)
		  {
		      (line->geometry + iv)->x *= scale_x;
		      (line->geometry + iv)->y *= scale_y;
		      if ((line->geometry + iv)->x < minx)
			  minx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y < miny)
			  miny = (line->geometry + iv)->y;
		      if ((line->geometry + iv)->x > maxx)
			  maxx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y > maxy)
			  maxy = (line->geometry + iv)->y;
		  }
		break;
	    case SPLE_POLYGON:
		polyg = (sple_polygon_ptr) (pGeo->geometry);
		line = polyg->external;
		for (iv = 0; iv < line->points; iv++)
		  {
		      (line->geometry + iv)->x *= scale_x;
		      (line->geometry + iv)->y *= scale_y;
		      if ((line->geometry + iv)->x < minx)
			  minx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y < miny)
			  miny = (line->geometry + iv)->y;
		      if ((line->geometry + iv)->x > maxx)
			  maxx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y > maxy)
			  maxy = (line->geometry + iv)->y;
		  }
		for (ib = 0; ib < polyg->borders; ib++)
		  {
		      line = *(polyg->internals + ib);
		      for (iv = 0; iv < line->points; iv++)
			{
			    (line->geometry + iv)->x *= scale_x;
			    (line->geometry + iv)->y *= scale_y;
			    if ((line->geometry + iv)->x < minx)
				minx = (line->geometry + iv)->x;
			    if ((line->geometry + iv)->y < miny)
				miny = (line->geometry + iv)->y;
			    if ((line->geometry + iv)->x > maxx)
				maxx = (line->geometry + iv)->x;
			    if ((line->geometry + iv)->y > maxy)
				maxy = (line->geometry + iv)->y;
			}
		  }
		break;
	    };
      }
    geom->minx += minx;
    geom->miny += miny;
    geom->maxx += maxx;
    geom->maxy += maxy;
/* builds the modified GEOMETRY to be passed back */
    sple_blob_encode (geom, result, size);	/* generating BLOB internal representation */
}

SPATIALITE_DECLARE void
sple_rotate_coords (sple_geometry_ptr geom, double angle,
		    unsigned char **result, int *size)
{
/* 
/ return a geometry that is the old old geometry with required rotation applied to coordinates
 */
    int ie;
    int ib;
    int iv;
    sple_generic_geometry_ptr pGeo;
    sple_point_ptr point;
    sple_polygon_ptr polyg;
    sple_linestring_ptr line;
    *size = 0;
    *result = NULL;
    double minx = DBL_MAX;
    double miny = DBL_MAX;
    double maxx = DBL_MIN;
    double maxy = DBL_MIN;
    double x;
    double y;
    double rad = angle * (M_PI / 180.0);
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		point = (sple_point_ptr) (pGeo->geometry);
		x = point->x;
		y = point->y;
		point->x = (x * cos (rad)) + (y * sin (rad));
		point->y = (y * cos (rad)) - (x * sin (rad));
		if (point->x < minx)
		    minx = point->x;
		if (point->y < miny)
		    miny = point->y;
		if (point->x > maxx)
		    maxx = point->x;
		if (point->y > maxy)
		    maxy = point->y;
		break;
	    case SPLE_LINESTRING:
		line = (sple_linestring_ptr) (pGeo->geometry);
		for (iv = 0; iv < line->points; iv++)
		  {
		      x = (line->geometry + iv)->x;
		      y = (line->geometry + iv)->y;
		      (line->geometry + iv)->x =
			  (x * cos (rad)) + (y * sin (rad));
		      (line->geometry + iv)->y =
			  (y * cos (rad)) - (x * sin (rad));
		      if ((line->geometry + iv)->x < minx)
			  minx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y < miny)
			  miny = (line->geometry + iv)->y;
		      if ((line->geometry + iv)->x > maxx)
			  maxx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y > maxy)
			  maxy = (line->geometry + iv)->y;
		  }
		break;
	    case SPLE_POLYGON:
		polyg = (sple_polygon_ptr) (pGeo->geometry);
		line = polyg->external;
		for (iv = 0; iv < line->points; iv++)
		  {
		      x = (line->geometry + iv)->x;
		      y = (line->geometry + iv)->y;
		      (line->geometry + iv)->x =
			  (x * cos (rad)) + (y * sin (rad));
		      (line->geometry + iv)->y =
			  (y * cos (rad)) - (x * sin (rad));
		      if ((line->geometry + iv)->x < minx)
			  minx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y < miny)
			  miny = (line->geometry + iv)->y;
		      if ((line->geometry + iv)->x > maxx)
			  maxx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y > maxy)
			  maxy = (line->geometry + iv)->y;
		  }
		for (ib = 0; ib < polyg->borders; ib++)
		  {
		      line = *(polyg->internals + ib);
		      for (iv = 0; iv < line->points; iv++)
			{
			    x = (line->geometry + iv)->x;
			    y = (line->geometry + iv)->y;
			    (line->geometry + iv)->x =
				(x * cos (rad)) + (y * sin (rad));
			    (line->geometry + iv)->y =
				(y * cos (rad)) - (x * sin (rad));
			    if ((line->geometry + iv)->x < minx)
				minx = (line->geometry + iv)->x;
			    if ((line->geometry + iv)->y < miny)
				miny = (line->geometry + iv)->y;
			    if ((line->geometry + iv)->x > maxx)
				maxx = (line->geometry + iv)->x;
			    if ((line->geometry + iv)->y > maxy)
				maxy = (line->geometry + iv)->y;
			}
		  }
		break;
	    };
      }
    geom->minx += minx;
    geom->miny += miny;
    geom->maxx += maxx;
    geom->maxy += maxy;
/* builds the modified GEOMETRY to be passed back */
    sple_blob_encode (geom, result, size);	/* generating BLOB internal representation */
}

SPATIALITE_DECLARE void
sple_reflect_coords (sple_geometry_ptr geom, int x_axis, int y_axis,
		     unsigned char **result, int *size)
{
/* 
/ return a geometry that is the old old geometry with required reflection applied to coordinates
 */
    int ie;
    int ib;
    int iv;
    sple_generic_geometry_ptr pGeo;
    sple_point_ptr point;
    sple_polygon_ptr polyg;
    sple_linestring_ptr line;
    *size = 0;
    *result = NULL;
    double minx = DBL_MAX;
    double miny = DBL_MAX;
    double maxx = DBL_MIN;
    double maxy = DBL_MIN;
    for (ie = 0; ie < geom->entities; ie++)
      {
	  pGeo = *(geom->geometries + ie);
	  switch (pGeo->type)
	    {
	    case SPLE_POINT:
		point = (sple_point_ptr) (pGeo->geometry);
		if (x_axis)
		    point->x *= -1.0;
		if (y_axis)
		    point->y *= -1.0;
		if (point->x < minx)
		    minx = point->x;
		if (point->y < miny)
		    miny = point->y;
		if (point->x > maxx)
		    maxx = point->x;
		if (point->y > maxy)
		    maxy = point->y;
		break;
	    case SPLE_LINESTRING:
		line = (sple_linestring_ptr) (pGeo->geometry);
		for (iv = 0; iv < line->points; iv++)
		  {
		      if (x_axis)
			  (line->geometry + iv)->x *= -1.0;
		      if (y_axis)
			  (line->geometry + iv)->y *= -1.0;
		      if ((line->geometry + iv)->x < minx)
			  minx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y < miny)
			  miny = (line->geometry + iv)->y;
		      if ((line->geometry + iv)->x > maxx)
			  maxx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y > maxy)
			  maxy = (line->geometry + iv)->y;
		  }
		break;
	    case SPLE_POLYGON:
		polyg = (sple_polygon_ptr) (pGeo->geometry);
		line = polyg->external;
		for (iv = 0; iv < line->points; iv++)
		  {
		      if (x_axis)
			  (line->geometry + iv)->x *= -1.0;
		      if (y_axis)
			  (line->geometry + iv)->y *= -1.0;
		      if ((line->geometry + iv)->x < minx)
			  minx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y < miny)
			  miny = (line->geometry + iv)->y;
		      if ((line->geometry + iv)->x > maxx)
			  maxx = (line->geometry + iv)->x;
		      if ((line->geometry + iv)->y > maxy)
			  maxy = (line->geometry + iv)->y;
		  }
		for (ib = 0; ib < polyg->borders; ib++)
		  {
		      line = *(polyg->internals + ib);
		      for (iv = 0; iv < line->points; iv++)
			{
			    if (x_axis)
				(line->geometry + iv)->x *= -1.0;
			    if (y_axis)
				(line->geometry + iv)->y *= -1.0;
			    if ((line->geometry + iv)->x < minx)
				minx = (line->geometry + iv)->x;
			    if ((line->geometry + iv)->y < miny)
				miny = (line->geometry + iv)->y;
			    if ((line->geometry + iv)->x > maxx)
				maxx = (line->geometry + iv)->x;
			    if ((line->geometry + iv)->y > maxy)
				maxy = (line->geometry + iv)->y;
			}
		  }
		break;
	    };
      }
    geom->minx += minx;
    geom->miny += miny;
    geom->maxx += maxx;
    geom->maxy += maxy;
/* builds the modified GEOMETRY to be passed back */
    sple_blob_encode (geom, result, size);	/* generating BLOB internal representation */
}
